#include "PhysFSOgre.h"
#include "PhysFS++.h"

#include <OgreArchiveFactory.h>
#include <OgreArchive.h>
#include <OgreArchiveManager.h>
#include <OgreDataStream.h>
#include <OgreString.h>


namespace PhysFS
{
  /** Datastream implementation for Ogre */
  class DataStream : public Ogre::DataStream
  {
  public:
    DataStream(const Ogre::String& filename);
    DataStream(const Ogre::String& name, const Ogre::String& filename);
    ~DataStream();

    size_t read(void* buf, size_t count);
    void skip(long count);
    void seek(size_t pos);
    size_t tell() const;
    bool eof() const;
    void close();

  private:
    PHYSFS_File* file;
  };

  DataStream::DataStream(const Ogre::String& filename)
  {
    file = PHYSFS_openRead(filename.c_str());
    if (!file)
      throw Exception(PHYSFS_getLastError());
    mSize = (size_t)PHYSFS_fileLength(file);
  }

  DataStream::DataStream(const Ogre::String& name, const Ogre::String& filename)
    : Ogre::DataStream(name)
  {
    file = PHYSFS_openRead(filename.c_str());
    if (!file)
      throw Exception(PHYSFS_getLastError());
    mSize = (size_t)PHYSFS_fileLength(file);
  }

  DataStream::~DataStream()
  {
    close();
  }

  size_t DataStream::read(void* buf, size_t count)
  {
    int64 read = PHYSFS_read(file, buf, 1, (PHYSFS_uint32)count);
    if (read == -1)
      throw Exception(PHYSFS_getLastError());
    return size_t(read);
  }

  void DataStream::seek(size_t pos)
  {
    if (!PHYSFS_seek(file, pos))
      throw Exception(PHYSFS_getLastError());
  }

  size_t DataStream::tell() const
  {
    return (size_t)PHYSFS_tell(file);
  }

  void DataStream::skip(long count)
  {
    size_t pos = tell() + count;
    seek(pos);
  }

  bool DataStream::eof() const
  {
    return PHYSFS_eof(file) != 0;
  }

  void DataStream::close()
  {
    if (file)
      PHYSFS_close(file);
    file = 0;
  }




  /** Archive implementation for Ogre */
  class Archive : public Ogre::Archive
  {
  public:
    Archive(const Ogre::String& name, const Ogre::String& type) : Ogre::Archive(name, type) { }

    /** PhysFS is case sensitive in general */
    bool isCaseSensitive() const { return true; }

    /** Nothing to load */
    void load() { }
    /** Nothing to unload */
    void unload() { }

    Ogre::DataStreamPtr open(const Ogre::String& filename) const;

    Ogre::StringVectorPtr list(bool recursive = true, bool dirs = false);
    Ogre::FileInfoListPtr listFileInfo(bool recursive = true, bool dirs = false);

    Ogre::StringVectorPtr find(const Ogre::String& pattern, bool recursive = true, bool dirs = false);

    bool exists(const Ogre::String& filename);
    
    Ogre::FileInfoListPtr findFileInfo(const Ogre::String& pattern, bool recursive = true, bool dirs = false);

  private:
    void listInfoRecursive(const Ogre::String& base, bool recursive, bool dirs, Ogre::FileInfoListPtr fileInfoList);
    void listRecursive(const Ogre::String& base, bool recursive, bool dirs, Ogre::StringVectorPtr fileList);
  };


  bool Archive::exists(const Ogre::String& filename)
  {
    return PhysFS::exists(mName + '/' + filename);
  }

  Ogre::DataStreamPtr Archive::open(const Ogre::String& filename) const
  {
    Ogre::String fullName = mName + '/' + filename;
    return Ogre::DataStreamPtr(new DataStream(filename, fullName));
  }


  void Archive::listInfoRecursive(const Ogre::String& base, bool recursive, bool dirs, Ogre::FileInfoListPtr fileInfoList)
  {
    Ogre::String baseDir = mName + '/' + base;
    StringVector files = enumerateFiles(baseDir);

    Ogre::FileInfo fileInfo;
    fileInfo.archive = this;
    fileInfo.path = base;
    fileInfo.compressedSize = 0;

    // iterate over all files and directories in the given directory
    for (StringVector::iterator it = files.begin(); it != files.end(); ++it)
    {
      fileInfo.basename = *it;
      fileInfo.filename = base + *it;
      if (isDirectory(*it))
      {
        if (dirs)
        {
          fileInfo.uncompressedSize = 0;
          fileInfoList->push_back(fileInfo);
        }
        if (recursive)
          listInfoRecursive(base+*it+'/', recursive, dirs, fileInfoList);
      }
      else
      {
        if (!dirs)
        {
          // get file size
          PHYSFS_File* file = PHYSFS_openRead((mName+'/'+fileInfo.filename).c_str());
          fileInfo.uncompressedSize = (size_t) PHYSFS_fileLength(file);
          PHYSFS_close(file);

          fileInfoList->push_back(fileInfo);
        }
      }
    }
  }

  void Archive::listRecursive(const Ogre::String& base, bool recursive, bool dirs, Ogre::StringVectorPtr fileList)
  {
    Ogre::String baseDir = mName + '/' + base;
    StringVector files = enumerateFiles(baseDir);

    // iterate over all files and directories in the given directory
    for (StringVector::iterator it = files.begin(); it != files.end(); ++it)
    {
      if (isDirectory(*it))
      {
        if (dirs)
          fileList->push_back(base + *it);
        if (recursive)
          listRecursive(base+*it+'/', recursive, dirs, fileList);
      }
      else
      {
        if (!dirs)
          fileList->push_back(base + *it);
      }
    }
  }


  Ogre::FileInfoListPtr Archive::listFileInfo(bool recursive, bool dirs)
  {
    Ogre::FileInfoListPtr fileInfoList (new Ogre::FileInfoList());
    listInfoRecursive("", recursive, dirs, fileInfoList);
    return fileInfoList;
  }

  Ogre::StringVectorPtr Archive::list(bool recursive, bool dirs)
  {
    Ogre::StringVectorPtr fileList (new Ogre::StringVector());
    listRecursive("", recursive, dirs, fileList);
    return fileList;
  }


  Ogre::StringVectorPtr Archive::find(const Ogre::String& pattern, bool recursive, bool dirs)
  {
    Ogre::StringVectorPtr fileList = list(recursive, dirs);
    Ogre::StringVectorPtr ret (new Ogre::StringVector());

    for (Ogre::StringVector::iterator it = fileList->begin(); it != fileList->end(); ++it)
    {
      if (Ogre::StringUtil::match(*it, pattern))
        ret->push_back(*it);
    }

    return ret;
  }

  Ogre::FileInfoListPtr Archive::findFileInfo(const Ogre::String& pattern, bool recursive, bool dirs)
  {
    Ogre::FileInfoListPtr fileList = listFileInfo(recursive, dirs);
    Ogre::FileInfoListPtr ret (new Ogre::FileInfoList());

    for (Ogre::FileInfoList::iterator it = fileList->begin(); it != fileList->end(); ++it)
    {
      if (Ogre::StringUtil::match(it->filename, pattern))
        ret->push_back(*it);
    }

    return ret;
  }


  /** ArchiveFactory for PhysFS */
  class ArchiveFactory : public Ogre::ArchiveFactory
  {
  public:
    /** ArchiveFactory typename: PhysFS */
    const Ogre::String& getType() const { static const Ogre::String type = "PhysFS"; return type; }

    Ogre::Archive* createInstance(const Ogre::String& name)
    {
      return new Archive(name, getType());
    }

    void destroyInstance(Ogre::Archive* instance)
    {
      delete instance;
    }


    static ArchiveFactory* getInstance()
    {
      static ArchiveFactory instance;
      return &instance;
    }

  private:
    ArchiveFactory() {}
    ArchiveFactory(const ArchiveFactory& o);
    ArchiveFactory& operator=(const ArchiveFactory& o);
  };


  void registerPhysFSToOgre()
  {
    Ogre::ArchiveManager::getSingleton().addArchiveFactory(ArchiveFactory::getInstance());
  }

}
